/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef HAVE_CONFIG_H
#include <config.h>
#endif

#include <artnet.h> 
#include <artnet/packets.h> 


#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>


#ifndef _defs_defined
#define _defs_defined

// first artnet specific stuff

// port to listen and send from
#define ARTNET_PORT 6454

#define ARTNET_STRING "Art-Net\0"

#define ARTNET_STRING_SIZE 8

#define ARTNET_VERSION 14
#define OEM_CODE 0xff

#define TTM_BEHAVIOUR_MASK 0x02
#define TTM_REPLY_MASK 0x01

#define PROGRAM_NO_CHANGE	0x7f
#define PROGRAM_DEFAULTS	0x00
#define PROGRAM_CHANGE_MASK	0x80

#define HIGH_NIBBLE 0xF0
#define LOW_NIBBLE 0x0F

#define STATUS_PROG_AUTH_MASK 0x30

#define PORT_STATUS_LPT_MODE	0x02
#define PORT_STATUS_SHORT		0x04
#define PORT_STATUS_ERROR		0x04
#define PORT_STATUS_DISABLED_MASK 0x08
#define PORT_STATUS_MERGE		0x08
#define PORT_STATUS_DMX_TEXT	0x10
#define PORT_STATUS_DMX_SIP		0x20
#define PORT_STATUS_DMX_TEST	0x40
#define PORT_STATUS_ACT_MASK 	0x80

#define PORT_DISABLE_MASK 0x01

#define MIN_PACKET_SIZE 10

#define MERGE_TIMEOUT_SECONDS 10

#define FIRMWARE_TIMEOUT_SECONDS 20

// non artnet specific
#define SA struct sockaddr
#define SI struct in_addr

#define TRUE 1
#define FALSE 0

#define min(a,b)    ((a) < (b) ? (a) : (b))
#define max(a,b)    ((a) > (b) ? (a) : (b))

/*
 * These are enums for fields in packets
 * Ordered by packets they appear in. Some of these appear in the public header file
 * artnet.h if they have to be used by the user.
 *
 * They are all in the form artnet_xxx_code in case they have to be made public some day
 *
 */

// ArtPollReply 
//-----------------------------------------------------------------------------

// the node report codes
typedef enum {
	ARTNET_RCDEBUG,
	ARTNET_RCPOWEROK,
	ARTNET_RCPOWERFAIL,
	ARTNET_RCSOCKETWR1,
	ARTNET_RCPARSEFAIL,
	ARTNET_RCUDPFAIL,
	ARTNET_RCSHNAMEOK,
	ARTNET_RCLONAMEOK,
	ARTNET_RCDMXERROR,
	ARTNET_RCDMXUDPFULL,
	ARTNET_RCDMXRXFULL,
	ARTNET_RCSWITCHERR,
	ARTNET_RCCONFIGERR,
	ARTNET_RCDMXSHORT,
	ARTNET_RCFIRMWAREFAIL,
	ARTNET_RCUSERFAIL
} artnet_node_report_code ;

// these define the types of node that can exist
// note it's different from artnet_node_type
typedef enum {
	STNODE = 0x00,
	STSERVER = 0x01,
	STMEDIA = 0x02,
	STROUTE = 0x03,
	STBACKUP = 0x04,
	STCONFIG = 0x05
} artnet_node_style_code;

// artnet_port_data in artnet.h

// ArtAddress
//-----------------------------------------------------------------------------

// artnet_port_command_code in artnet.h

// ArtFirmwareMaster
//-----------------------------------------------------------------------------

// defines contents of the firmware packet
typedef enum  {
	ARTNET_FIRMWARE_FIRMFIRST = 0x00,
	ARTNET_FIRMWARE_FIRMCONT = 0x01,	
	ARTNET_FIRMWARE_FIRMLAST = 0x02,	
	ARTNET_FIRMWARE_UBEAFIRST = 0x03,	
	ARTNET_FIRMWARE_UBEACONT = 0x04,	
	ARTNET_FIRMWARE_UBEALAST = 0x05,	
} artnet_firmware_type_code ;

// ArtFirmwareReply
//-----------------------------------------------------------------------------

// artnet_firmware_status_code in artnet.h



// End packet enums
//-----------------------------------------------------------------------------


/*
 * Data structures to manage callbacks
 */

// packet callbacks have a function and some user data
typedef struct {
	int (*fh)(artnet_node n, void *p, void *data) ;
	void *data;	
} callback_t ;

/*
 * the dmx callback is triggered when a dmx packet arrives
 */
typedef struct {
	int (*fh)(artnet_node n, int portid, void *data) ;
	void *data;	
} dmx_callback_t ;

/*
 * firmware callback is trigger when a firmware recieve has been completed sucessfully
 */
typedef struct {
	int (*fh)(artnet_node n, int ubea, uint16_t *data, int length, void *d) ;
	void *data;	
} firmware_callback_t ;


// struct to store callbacks
typedef struct {
	callback_t recv;
	callback_t send;
	callback_t poll;
	callback_t reply;
	callback_t dmx;
	callback_t address;
	callback_t input;
	callback_t todrequest;
	callback_t toddata;
	callback_t ipprog;
	callback_t firmware ;
	callback_t firmware_reply;
	dmx_callback_t dmx_c ;
	firmware_callback_t firmware_c ;
} node_callbacks_t ;


// End callback structures
//-----------------------------------------------------------------------------

/*
 * Begin port structures
 */

// first a generic port
typedef struct {
	uint8_t addr ;					// the port address
	uint8_t default_addr ;			// the address set by the hardware
	uint8_t net_ctl ;				// if the port address is under network control
	uint8_t status ;				// status of the port
} port_t ;

/**
 * struct to represent an input port
 * input ports need to keep track of sequence numbers
 * (this doesn't seem to be used though)
 */
typedef struct {
	port_t port;
	uint8_t seq ;
} input_port_t ;


/**
 * For output ports we need to track if they merge in HTP or LTP modes
 */
typedef enum {
	ARTNET_MERGE_HTP,
	ARTNET_MERGE_LTP
} merge_t ;


/**
 * struct to represent an output port
 * 
 * output ports can merge data from two sources in either HTP 
 * (highest takes precedence) or LPT (lowest takes precedence) mode
 * 
 * we need to store:
 *   o The data from each source
 *   o The ip of the source
 *   o The time the data was recv'ed
 *
 */
typedef struct {
	port_t port;
	int	length;								// the length of the data THAT HAS CHANGED since the last dmx packet
	uint8_t enabled ;						//  true if the port has had it's address set, this is internal only, it's not used by the ArtNet protocol, otherwise the node keeps picking up packets for the 0x00 port
	uint8_t	data[ARTNET_MAX_DMX] ;			// output data
	// for merging
	merge_t merge_mode;
	uint8_t dataA[ARTNET_MAX_DMX] ;
	uint8_t dataB[ARTNET_MAX_DMX] ;
	time_t timeA ;
	time_t timeB ;
	SI ipA;
	SI ipB;
} output_port_t ;

// use defines to hide the inner structures
#define port_addr port.addr
#define port_default_addr port.default_addr
#define port_net_ctl port.net_ctl
#define port_status port.status



// End port structures
//-----------------------------------------------------------------------------

/*
 * We use a linked list to keep track of nodes on the network
 * Here be the structures
 */

/* firstly we have the potential to do a firmware transfer to any node,
 * this struct keeps the information such as how much data has been transfered
 * and the address of the peer. It's also used for receiving firmware
 */
typedef struct {
	uint16_t	*data ;
	int			bytes_current ;
	int 		bytes_total ;
	struct 		in_addr peer;
	int			ubea ;
	time_t 		last_time ;
	int			expected_block ;
	int 		(*callback)(artnet_node n, artnet_firmware_status_code code, void *d);
	void 		*user_data;
} firmware_transfer_t ;

/*
 * The node entry in the LL. It contains the public entry, as well as some stuff
 * we don't want public like firmware
 */
typedef struct node_entry_private_s {
	artnet_node_entry_t pub ;
	struct node_entry_private_s *next;
	firmware_transfer_t firmware ;
	SI ip ;								// don't rely on the ip address that the node
										// sends, they could be faking it. This is the ip that
										// the pollreply was send from
										// besides it's annoying to deal with a uint8_t[4]
} node_entry_private_t ;

/**
 * The node list stores a pointer to the first,last and current 
 * entries.
 *
 */
typedef struct {
	node_entry_private_t *first;
	node_entry_private_t *current;
	node_entry_private_t *last;
	int length;	
} node_list_t ;


// End node list structures
//-----------------------------------------------------------------------------


// the status of the node
typedef enum {
	ARTNET_OFF,
	ARTNET_STANDBY,
	ARTNET_ON
} node_status_t ;


// struct to hold the state of the node
typedef struct {
	artnet_node_type node_type;
	node_status_t mode;
	SI reply_addr;
	SI ip_addr;
	SI bcast_addr ;
	uint8_t hw_addr[6] ;
	int send_apr_on_change;
	char short_name[ARTNET_SHORT_NAME_LENGTH] ;
	char long_name[ARTNET_LONG_NAME_LENGTH] ;
	char report[ARTNET_REPORT_LENGTH] ;
	int ar_count ;
	uint8_t subnet ;
	uint8_t default_subnet ;
	uint8_t subnet_net_ctl ;
	int verbose ;

	artnet_node_report_code report_code;
} node_state_t ;


/**
 * The main node structure
 */
typedef struct {
	int sd[2] ;									// the two sockets
	node_state_t state;							// the state struct
	node_callbacks_t callbacks;					// the callbacks struct
	struct ports_s {
		uint8_t	types[ARTNET_MAX_PORTS] ;		// type of port
		input_port_t in[ARTNET_MAX_PORTS] ;		// input ports
		output_port_t out[ARTNET_MAX_PORTS] ;	// output ports
	} ports;
	artnet_reply_t ar_temp ;					// buffered artpoll reply packet
	node_list_t node_list ;						// node list
	firmware_transfer_t firmware ;

} artnet_node_t ;


typedef artnet_node_t *node ;

/*
 * Function definitions follow
 */

// exported from artnet.c
void build_art_poll_reply(node n) ;
node_entry_private_t *find_private_entry( node n, artnet_node_entry e)  ;
void check_timeouts(node n) ;
node_entry_private_t *find_entry_from_ip(node_list_t *nl, SI ip ) ;


// exported from misc.c
void err_warn(const char * fmt , ... ) ;
void err_fatal(const char *fmt, ... ) ;
void *Malloc(int len);
void *Calloc(size_t nmemb, int len);
int16_t bytes_to_short(uint8_t h, uint8_t l) ;
int32_t nbytes_to_32(uint8_t bytes[4]) ;
void int_to_bytes(int data, uint8_t *bytes) ;
int32_t nbytes_to_h32(uint8_t bytes[4]) ;
uint8_t short_get_low_byte(int16_t i) ;
uint8_t short_get_high_byte(int16_t i) ;


// exported from network.c
ssize_t net_recv(node n, artnet_packet  p, int block) ;
int net_send(node n, artnet_packet p) ;
int net_set_non_block(node n) ;
int net_init(node n, const char *ip) ;
int net_start(node n) ;
int net_close(node n) ;


// exported from receive.c
int handle(node n, artnet_packet p ) ;
int16_t get_type(artnet_packet p ) ;
void reset_firmware_upload(node n) ;


// exported from transmit.c
int send_poll_reply(node n, int reply) ;
int send_tod_data(node n) ;
int send_firmware_reply(node n, in_addr_t ip, artnet_firmware_status_code code) ;
#endif
